#pragma once
// #include <math.h>
#include <assert.h>
#include <stdio.h>

template <uint FPL>
class fixedpoint {
protected:
	typedef fixedpoint self;

	uint value;

public:
	inline fixedpoint(const uint& v, const uint& s) {
		assert(s < FPL);
		value = (v * FPL) + s;
	}

	inline fixedpoint(const uint& v = 0) {
		value = v * FPL;
	}

	inline fixedpoint(const int& v) {
		assert(v >= 0);
		value = v * FPL;
	}

	inline fixedpoint(const float& v) {
		value = (uint)(((float)v*FPL) + 0.5);
	}

	inline fixedpoint(const double& v) {
		value = (uint)(((double)v*FPL) + 0.5);
	}

	inline fixedpoint(const fixedpoint* const src) : fixedpoint(*src) { }

	// copy constructor
	inline fixedpoint(const fixedpoint& src) : value(src.value) { }

	// move constructor
	inline fixedpoint(fixedpoint&& src) : value(src.value) { }
	
	// destructor
	inline ~fixedpoint() { };

	// copy assignement operator
	inline fixedpoint& operator=(const fixedpoint& src) {
		if (&src != this) {
			this->value = src.value;
		}
		return *this;
	}

	// move assignement operator
	inline fixedpoint& operator=(fixedpoint&& src)	{
		if (&src != this) {
			this->value = src.value;
			src.value = 0;
		}
		return *this;
	}

	inline fixedpoint& operator=(const int& v) {
		assert(v >= 0);
		this->value = (uint) v;
		return *this;
	}

	inline fixedpoint& operator=(const uint& v) {
		this->value = v;
		return *this;
	}

	inline fixedpoint& operator=(const float& v) {
		this->value = (uint)(((float) v*FPL) + 0.5);
		return *this;
	}

	inline fixedpoint& operator=(const double& v) {
		this->value = (uint)(((double) v*FPL) + 0.5);
		return *this;
	}

	inline fixedpoint& operator++() {
		++this->value;
		return *this;
	}
	inline fixedpoint operator++(int) {
		fixedpoint tmp(*this); // copy
		operator++(); // pre-increment
		return tmp;   // return old value
	}
	inline fixedpoint& operator--() {
		--this->value;
		return *this;
	}
	inline fixedpoint operator--(int) {
		fixedpoint tmp(*this); // copy
		operator--(); // pre-increment
		return tmp;   // return old value
	}

	inline fixedpoint operator+(const fixedpoint& rhs) {
		fixedpoint result;
		result.value = value + rhs.value;
		return result;
	}

	inline fixedpoint operator+=(const fixedpoint& rhs) {
		value += rhs.value;
		return *this;
	}

	inline bool operator==(const fixedpoint& rhs) const { return this->value == rhs.value ; }
	inline bool operator<(const fixedpoint& rhs) const {	return (this->value < rhs.value);	}

	inline bool operator!=(const fixedpoint& rhs) const { return !(*this == rhs); }
	inline bool operator>(const fixedpoint& rhs) const { return (rhs < *this); }
	inline bool operator<=(const fixedpoint& rhs) const { return !(*this > rhs); }
	inline bool operator>=(const fixedpoint& rhs) const { return !(*this < rhs); }

	void print() const {
		printf("%02d.%02d\n", value / FPL, value % FPL);
	}

};

